package es.unican.docencia.y2011.eda.listaEnlazada.objetos.generico;

/**
 * Clase lista enlazada (version orientada a objetos y generica).
 * @author Pablo Sanchez (p.sanchez@unican.es)
 * 		   http://personales.unican.es/sanchezbp	
 */
public class ListaEnlazada<Elemento extends IClave<C>,C> {
	
	// La lista enlazada orientada a objetos, por motivos de estetica, 
	// es un envoltorio (wrapper) sobre la lista no orientada a objetos
	// ademas, los mtodos se hacen dinamicos, lo que permite la vinculacion
	// dinamica (que no se usa en esta practica)
	
	// La filosofia cambia ligeramente, porque ahora la clase ha de responder a 
	// metodos
	
	// Como no queremos que nadie nos ande tocando los nodos, lo declaramos como
	// clase interna. Nodo representa ahora nuestra lista estatica
	protected class Nodo {
		protected Elemento dato = null;
		protected Nodo sig = null;
	} // Nodo
	
	// Variable que es envuelta por la clase (wrapped element)
	protected Nodo lista = null;
	
	// Dado que estamos usando orientacion a objetos , hay que hacer algun 
	// constructor, aunque dependiendo del lenguaje esto no hace falta.
	ListaEnlazada() {
		this.lista = null;
	} // ListaEnlazada 
	
	// Implementamos ahora las operaciones requeridas de forma estatica
		
	//-------------------------------------------------------------------//
	// Operaciones publicas sobre la clase								 //  
	//-------------------------------------------------------------------//
	
	/**
	 * Metodo que devuelve una lista vacia. Su intencin es que sea el generador
	 * del tipo, por l que hay que dejarlo esttico. Es decir, este metodo debe 
	 * existir antes de que cualquier objeto sea creado    
	 * @return Una lista vacia
	 */
	 public static <T extends IClave<C>,C> ListaEnlazada<T,C> listaVacia() {
		return new ListaEnlazada<T,C>();
	} // 
	
	/**
	 * Devuelve verdadero si la lista esta vacia, es decir, sino contiene 
	 * ningun elemento. Devuelve falso en caso contrario
	 * @return verdadero Si la lista esta vacia. Falso en caso contrario
	 */
	// FUNCION esVacia(l : ListaEnlazadaChinos) : Booleano;
	public boolean esVacia() {
		return (this.lista == null);
	} // esVacia
	
	/**
	 * Inserta un elemento en la lista enlazada. 
	 * NOTA: No se comprueba que est repetido 
	 * @param l La lista enlazada donde se inserta el elemento
	 * @return La lista enlazada l con el elemento insertado 
	 */
	// PROCEDIMIENTO inserta(REF l : ListaEnlazadaChinos, c : Chinos)
	// NOTA: Observad como el paso por referencia es simulado mediante la 
	//       devolucin de la lista posiblemente modificada  
	public ListaEnlazada<Elemento,C> inserta(Elemento elemento) {
		
		Nodo aux = new Nodo();
		aux.dato   = elemento;
		aux.sig    = this.lista;
		this.lista = aux;
	
		// Esto es convecion de Java de cara a la galera. Lo devuelva o no, 
		// la lista se ha modificado
		return this;
	} // inserta
	
	/**
	 * Borra un elemento de nuestra lista, concretamente el primero que se encuentra con la clave
	 * dada. Si no se encuentra el elemento, no se hace nada 
	 * @param clave La clave del elemento a borrar
	 * @return La lista enlazada l con el elemento borrado
	 */
	// PROC borra(REF l : ListaEnlazadaAlumnos, clave : Clave)
	public ListaEnlazada<Elemento,C> borra(C clave) {

		// Si est vaca, la dejamos vaca
		if (this.lista != null) {
			// Obviamente {P:(l != null)} se satisface en este punto y puedo dereferenciar el 
			// puntero sin problemas
			if (this.lista.dato.getClave().equals(clave)) {
				// Tengo que borrar el primer elemento

				//-----------------------------------------------------------------------
				// Esta lnea en Java no es necesaria, porque Java, para bien o para mal
				// recoge nuestra basura
				//-----------------------------------------------------------------------
				// ListaEnlazadaChinos aux = l; 
				
				// Avanzamos un elemento. Nadie apunta ahora la primer elemento de la lista, 
				// por lo que el recolector automtico de basura (garbage collector) algn da
				// liberar ese espacio de memoria
				//  OJO ! Si modificamos l, estamos cambiando una variable local y por tanto 
				// el cambio no se refleja en lista
				// System.out.println("Borrando el primero de la lista");
				this.lista = this.lista.sig;
				
				//-------------------------------------------------------------------------------
				// Si no hubiese recolector automtico de basura tendramos que aadir una linea
				// como la de abajo
				//-------------------------------------------------------------------------------
				// free(aux);
			} else {
				// Itero mientras me queden elementos por analizar y no haya encontrado al chino
				// Itero desde una posicin ms atrasada de la que analizo
				Nodo aux = this.lista;
				// System.out.println("Apunto a " + aux.dato.getClave());
				while ((aux.sig != null) && (!aux.sig.dato.getClave().equals(clave))) {
					aux = aux.sig;
//					if (aux.sig == null ) { 
//						System.out.println("Estoy al final de la lista");
//					} else {
//						System.out.println("Avanzo y apunto a " + aux.dato.getClave());
//					}
				} // while
				// En este punto o lo he encontrado o he llegado al final de la lista
				// Dicho predicado se asegura mediante la negacin de la condicin del bucle
				// Si lo he encontrado he de borrarlo, sino no hago nada
				if (aux.sig != null) {
					// Simplemente avanzo el puntero
					// Si Java no tuviese recoleccion automatica de basura, tendra que hacer lo mismo
					// que para el caso anterior, si deseo liberar el espacio en memoria ocupado por 
					// el nodo.
					aux.sig = aux.sig.sig;
				} // if
			} // if
		} // if
		
		// Esto es convecin de Java de cara a la galera. lo devuelva o no, 
		// la lista se ha modificado
		return this;
	} // borra
	
	/**
	 * Devuelve el primer elemento almacenado en la lista cuya clave coincide con la solicitada 
	 * @param clave La clave del elemento buscado
	 * @return Un puntero al primer elemento cuyo nombre coincide con el indicado, nulo si tal elemento
	 *         no se encuentra en nuestra lista
	 */
	// FUN busca(l : ListaEnlazadaAlumnos, n : Clave) : PUNTERO A Chino
	public Elemento busca(C clave) {
		// Simplemente itero por la lista
		Elemento result = null;
		Nodo l = buscaNodo(clave);
		// Si lo he encontrado, lo devuelvo 
		if (l != null) {
			result = l.dato; 
		} // if

		// Esto es convecin de Java de cara a la galera. lo devuelva o no, 
		// la lista se ha modificado
		return result;
	} //  busca

	/**
	 * Cambia el primer elemento cuya clave es igual a la que se pasa como 
	 * parametro por el valor de nuevoDato 
	 * @param clave Clave del elemento a modificar
	 * @param nuevoDato Dato que sustituye al elemento modificado 
	 * @return La lista actualizada 
	 */
	// PROC actualiza(l : ListaEnlazada, clave : Clave, nuevo : Chino);
	public ListaEnlazada<Elemento,C> actualiza(C clave, Elemento nuevoDato) {
		
		Nodo nodo = buscaNodo(clave);
		if (nodo != null) {
			nodo.dato = nuevoDato;
		} // if
		
		return this;
	} // actualiza
		
	//-----------------------------------------------------------------------------------//
	// 						Funciones auxiliares (helper functions)						 //
	//-----------------------------------------------------------------------------------//	
	
	/**
	 * Devuelve una referencia al nodo que contiene al primer elemento cuyo nombre se pasa como 
	 * parametro  
	 * @param clave Nombre de la clave que buscamos
	 * @return Referencia al nodo que contiene el primer elemento buscado si este existe, nulo 
	 * 		   en caso contrario	 
	 */
	protected Nodo buscaNodo(C clave) {
		
		Nodo result = null;
		
		Nodo l = this.lista; 
		
		while ((l != null) && (!l.dato.getClave().equals(clave))) {
			l = l.sig;
		} // while
		
		if (l != null) {
			result = l;
		} // if
		
		return result;
	} // buscaNodo
	
	//-----------------------------------------------------------------------------------//
	// 						Aburrida y soporfera entrada/salida						 //
	//-----------------------------------------------------------------------------------//	
	
	/**
	 * Devuelve una cadena con los nombre de todos los chinos serliazados por orden de insercion
	 * inverso, es decir, el ultimo chino insertado es el primero que aparece
	 * @return La mencionada cadena. 
	 */
	public String toString() {
		
		String result = "";
		
		Nodo l = this.lista;
		
		while(l != null) {
			// result = result + l.dato.toString() + "/" + l.sig + "; ";
			result = result + l.dato.toString() + "; ";
			l = l.sig;
		} //  while
		
		return result;
	} // toString
	
} // ListaEnlazada<Elemento,C>
